:- consult(cosim_semantics).
:- consult(opt_cosim_semantics).
:- consult(example_scenarios).

test(portsInSameTimeStamp):-
	\+portsInSameTimeStamp([port(p1,_), port(p2, _)], [pstate(p1, defined, t),pstate(p2, undefined, tH)], _),
	portsInSameTimeStamp([port(p1,_), port(p2, _)], [pstate(p1, defined, t),pstate(p2, undefined, t)], t).

test(executeSetIn):-
  scenario_example_feedback(FMUs, Connections),
  inStateT0(FMUs, StateBefore),
  StateAfter=[fstate(a, t, [pstate(u1, undefined, t)], [pstate(y1, defined, t)]), fstate(b, t, [pstate(u2, defined, _3026)], [pstate(y2, undefined, t)])],
  executeOp(getOut(a, y1), FMUs, _, StateBefore, SAUX),
        executeOp(setIn(b, u2), FMUs, Connections, SAUX, StateAfter).

test(getOut):-
  FMUs=[fmu(a, [port(u1,  delayed)],
              [port(y1,[u1])])],
  StateBefore=[fstate(a, t,
                [pstate(u1, defined,t)],
                [pstate(y1,undefined,t)])],
  StateAfter=[fstate(a, t,
                [pstate(u1, defined,t)],
                [pstate(y1, defined,t)])],
  executeOp(getOut(a, y1), FMUs, _,
            StateBefore, StateAfter).

test(doStep):-
  FMUs=[fmu(a, [port(u1,  reactive)],
              [port(y1,[u1])])],
  StateBefore=[fstate(a, t,
                [pstate(u1,defined,tH)],
                [pstate(y1,defined, t)])],
  StateAfter=[fstate(a, tH,
                [pstate(u1,defined,tH)],
                [pstate(y1,undefined,tH)])],
  executeOp(doStep(a), FMUs, _,
              StateBefore, StateAfter).

test(master):-
  scenario_example_feedback(FMUs, Connections),
  Init=[getOut(a, y1), setIn(b, u2), getOut(b, y2),
        setIn(a, u1)],
  Step=[doStep(a), getOut(a, y1), setIn(b, u2),
        doStep(b), getOut(b, y2), setIn(a, u1)],
  isMasterAlgorithm(Init, Step, FMUs, Connections).

test(generate_constraint_graph):-
	Graph = [
						getOut(a,y1) - [setIn(b,u2)],
						doStep(a) - [getOut(a,y1)],
						setIn(b,u2) - [],
						doStep(b) - [setIn(b,u2)]
					],
	scenario_example_termination(FMUs, Connections),
	generate_constraint_graph_step(FMUs, Connections, Graph).

test(generate_constraint_graph_init):-
	Graph = [
		getOut(a, y1)-[setIn(b, u2)],
		setIn(b, u2)-[]
		],
	scenario_example_termination(FMUs, Connections),
	generate_constraint_graph_init(FMUs, Connections, Graph).

test(ancestors):-
	Graph = [1-[2,5,3], 2-[4], 3-[4], 5-[4], 4-[]],
	\+ancestors(2, Graph, []),
	ancestors(4, Graph, [2,3,5]),
	ancestors(1, Graph, []),
	ancestors(2, Graph, [1]).

test(all_sources):-
	Graph = [1-[2,5,3], 2-[4], 3-[4], 5-[4], 4-[], 6-[4]],
	Sources=[1,6],
	find_all_sources(Graph, Sources).

test(all_top_sort):-
	Graph = [1-[2,5,3], 2-[4], 3-[4], 5-[4], 4-[]],
	OpsOut = [[1], [2, 3, 5], [4]],
	all_top_sort(Graph, OpsOut).

test(schedule_init_complex):-
	example_scenario_complex(FMUs, Connections),
	generate_init_opt_bundled(FMUs, Connections, _).

test(schedule_step_complex):-
	example_scenario_complex(FMUs, Connections),
	generate_step_opt_bundled(FMUs, Connections, _).

test(terminating, X):-
	FMUs = [fmu(b,[port(u1,delayed),port(u2,delayed)],[port(y1,[]),port(y2,[u1]),port(y3,[]),port(y4,[])]),fmu(a,[port(u1,delayed),port(u2,delayed),port(u3,reactive),port(u4,reactive)],[port(y1,[u1,u3]),port(y2,[])])],
	Connections=[connect(a,y2,b,u2),connect(b,y1,a,u1),connect(b,y2,a,u2),connect(a,y1,b,u1),connect(b,y3,a,u3),connect(b,y4,a,u4)],
	generate_init_opt_bundled(FMUs,Connections,X).

test(non_terminating, Graph, Ops):-
	FMUs = [fmu(b,[port(bu1,delayed),port(bu2,delayed)],[port(by1,[]),port(by2,[bu1]),port(by3,[]),port(by4,[])]),
				fmu(a,[port(au1,delayed),port(au2,delayed),port(au3,reactive),port(au4,reactive)],[port(ay1,[au1,au3]),port(ay2,[])])],
	Connections=[connect(a,ay2,b,bu2),connect(b,by1,a,au1),connect(b,by2,a,bu2),connect(a,ay1,b,bu1),connect(b,by3,a,au3),connect(b,by4,a,au4)],
	generate_constraint_graph_init(FMUs, Connections, Graph),
	all_top_sort(Graph, Ops).

test(all):-
	test(portsInSameTimeStamp), !,
	write('portsInSameTimeStamp passed\n'),
	test(executeSetIn), !,
	write('executeSetIn passed\n'),
	test(getOut), !,
	write('getOut passed\n'),
	test(doStep), !,
	write('doStep passed\n'),
	test(master), !,
	write('master passed\n'),
	test(generate_constraint_graph), !,
	write('generate_constraint_graph passed\n'),
	test(generate_constraint_graph_init), !,
	write('generate_constraint_graph_init passed\n'),
	test(ancestors), !,
	write('ancestors passed\n'),
	test(all_sources), !,
	write('all_sources passed\n'),
	test(all_top_sort), !,
	write('all_top_sort passed\n'),
	test(schedule_init_complex), !,
	write('schedule_init_complex passed\n'),
	test(schedule_step_complex), !,
	write('schedule_step_complex passed\n').
